/*

  sshttyflags.h
  
  Author: Sami Lehtinen <sjl@ssh.com>

  Copyright (C) 1999-2000 SSH Communications Security Corp, Helsinki, Finland
  All rights reserved.

  Prototypes for public stty-mode manipulation functions.
  
*/

/*
 * $Id: sshttyflags.h,v 1.5 2000/05/05 14:13:01 tri Exp $
 * $Log: sshttyflags.h,v $
 * $EndLog$
 */

#ifndef SSHTTYFLAGS_H
#define SSHTTYFLAGS_H

void ssh_decode_tty_flags(int fd, unsigned char *buf, size_t buf_len);
void ssh_encode_tty_flags(int fd, unsigned char **buf, size_t *buf_len);
void ssh1_encode_tty_flags(int fd, unsigned char **buf, size_t *buf_len);

#endif /* SSHTTYFLAGS_H */
