#ifdef SSH_FSM_PROTOTYPES
#define SSH_FSM_DECL(x,y,z) SSH_FSM_STEP(z);
#else
#define SSH_FSM_DECL SSH_FSM_STATE
#endif /* SSH_FSM_PROTOTYPES */

SSH_FSM_DECL("pager_buf_to_out", "Transfer data from buffer to output",
             sftp_pager_buf_to_out)
SSH_FSM_DECL("pager_empty_input", "Slurp all the input", sftp_pager_empty_input)
SSH_FSM_DECL("pager_write_prompt", "Write the prompt", sftp_pager_write_prompt)
SSH_FSM_DECL("pager_read_a_char", "Read one character", sftp_pager_read_a_char)
SSH_FSM_DECL("pager_reset", "Reset the output state", sftp_pager_reset)
SSH_FSM_DECL("pager_finish", "Finish the Pager thread", sftp_pager_finish)    
SSH_FSM_DECL("pager_input_availability_notify", "Input availability notifier thread (only state)", sftp_pager_input_availability_notify)

#undef SSH_FSM_DECL
#undef SSH_FSM_PROTOTYPES
