/* THIS FILE IS AUTOMATICALLY GENERATED, DO NOT EDIT! */
/* Copyright (c) 1998 SSH Communications Security Ltd. */

/* This file was generated by the SSH distribution system, and should
   not normally be edited.  However, editing this file by a customer may
   sometimes make sense to disable some features.

   This file should never be edited by anyone inside SSH
   except for testing purposes; distribution configurations should
   be defined in the distributions file. */

#ifndef SSHDISTDEFS_H
#define SSHDISTDEFS_H

#define SSHDIST_CRYPT_DL
#define SSHDIST_UTIL_FSM
#define SSHDIST_CRYPT_DLKEX
#define SSHDIST_CRYPT_RIPEMD160
#define SSHDIST_CRYPT_TWOFISH
#define SSHDIST_CRYPT_GENMAC
#define SSHDIST_MEDIA_ETHERNET
#define SSHDIST_SESSION
#define SSHDIST_PGP
#define SSHDIST_UTIL_REALPATH
#define SSHDIST_AGENT
#define SSHDIST_CRYPT_GENHASH
#define SSHDIST_PROTO
#define SSHDIST_UTIL_STRUTIL
#define SSHDIST_IPSEC_FILTER_OPTIMIZER
#define SSHDIST_CRYPT_COMPRESS
#define SSHDIST_HTTP
#define SSHDIST_UTIL_NET
#define SSHDIST_UTIL_CORE
#define SSHDIST_UTIL_STREAM
#define SSHDIST_MATH
#define SSHDIST_IPSECAPPAPI
#define SSHDIST_CRYPT_BLOWFISH
#define SSHDIST_UTIL_FILEIO
#define SSHDIST_UTIL_BUFFER
#define SSHDIST_UTIL_CRC32
#define SSHDIST_SCM
#define SSHDIST_CRYPT_ARCFOUR
#define SSHDIST_IPSEC_CFG_MODE
#define SSHDIST_CRYPT_DSA
#define SSHDIST_UTIL_MATCH
#define SSHDIST_UTIL_GETOPT
#define SSHDIST_UTIL_FILEXFER
#define SSHDIST_UTIL_STREAM_LOCAL
#define SSHDIST_FILEXFER
#define SSHDIST_UTIL_FILEBUFFER
#define SSHDIST_EXTKEY_SOFT_ACCELERATOR_PROV
#define SSHDIST_SSH2
#define SSHDIST_UTIL_PACKETSTREAM
#define SSHDIST_UTIL_TCP
#define SSHDIST_ZLIB
#define SSHDIST_UTIL_TESTUTIL
#define SSHDIST_SMTP
#define SSHDIST_UTIL_COMMON
#define SSHDIST_UTIL_MPAUX
#define SSHDIST_PLATFORM_UNIX
#define SSHDIST_UTIL_MISC
#define SSHDIST_UTIL_MAPPING
#define SSHDIST_CRYPT_DES
#define SSHDIST_UTIL_SIGNALS
#define SSHDIST_UTIL_URL
#define SSHDIST_SCRLOCK
#define SSHDIST_CRYPT
#define SSHDIST_IPSEC_AUTH_CERT
#define SSHDIST_CRYPT_MD5
#define SSHDIST_CRYPT_KEYBLOB
#define SSHDIST_CRYPT_GENPKCS
#define SSHDIST_UTIL
#define SSHDIST_APPUTIL
#define SSHDIST_UTIL_UTF8
#define SSHDIST_CRYPT_SHA
#define SSHDIST_UTIL_STREAM_FILTER
#define SSHDIST_IPSEC_CRASH_RECOVERY
#define SSHDIST_UTIL_OLDADT
#define SSHDIST_UTIL_ADT
#define SSHDIST_UTIL_SYSUTIL
#define SSHDIST_HS
#define SSHDIST_UTIL_ELOOP
#define SSHDIST_UTIL_BASE64
#define SSHDIST_CRYPT_SSHMACS
#define SSHDIST_READLINE
#define SSHDIST_UTIL_UDP
#define SSHDIST_UTIL_FILELOCK
#define SSHDIST_UTIL_DLLIST
#define SSHDIST_CRYPT_HMAC
#define SSHDIST_EXTERNALKEY
#define SSHDIST_CRYPT_GENCIPH
#define SSHDIST_UTIL_STRING
#define SSHDIST_UTIL_STREAM_PAIR
#define SSHDIST_CRYPT_DH
#define SSHDIST_UTIL_TIMEMEASURE

#endif /* SSHDISTDEFS_H */
