/*

  sshadt_bag.h

  Author: Antti Huima <huima@ssh.fi>

  Copyright (c) 1999, 2000 SSH Communications Security, Finland
  All rights reserved.

  Created Fri Sep 24 13:24:19 1999.

  */

#ifndef SSH_ADT_BAG_H_INCLUDED
#define SSH_ADT_BAG_H_INCLUDED

extern const SshADTContainerType ssh_adt_type_bag;

#define SSH_ADT_BAG (ssh_adt_type_bag)

#endif /* SSH_ADT_BAG_H_INCLUDED */
