/*

  sshgetcwd.h

  Author:
        Sami Lehtinen <sjl@ssh.com>

  Copyright (C) 1999-2000 SSH Communications Security Corp, Helsinki, Finland
  All rights reserved.

*/

#ifndef SSHGETCWD_H
#define SSHGETCWD_H

/* An easy interface for getcwd(). Returns a ssh_xmalloceted buffer,
   that contains the current working directory, or NULL on error. */
char *ssh_getcwd(void);

#endif /* SSHGETCWD_H */
