/*

  sshrgf.h

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (c) 1999-2000 SSH Communications Security, Finland
  All rights reserved.

  Created Thu Dec  9 23:42:00 1999.

  */

/*
  Redundancy generating functions for public key cryptosystems.
 */

#ifndef SSHRGF_H
#define SSHRGF_H

SshRGFHash
ssh_rgf_hash_allocate_internal(const SshRGFDef *rgf_def);

/* Definition structures. */

extern const SshRGFDef ssh_rgf_pkcs1_sha1_def;
extern const SshRGFDef ssh_rgf_pkcs1_md5_def;
extern const SshRGFDef ssh_rgf_pkcs1_none_def;

extern const SshRGFDef ssh_rgf_pkcs1v2_sha1_def;
extern const SshRGFDef ssh_rgf_pkcs1v2_md5_def;
extern const SshRGFDef ssh_rgf_pkcs1v2_none_def;

extern const SshRGFDef ssh_rgf_std_sha1_def;
extern const SshRGFDef ssh_rgf_std_md5_def;

extern const SshRGFDef ssh_rgf_dummy_def;

#endif /* SSHRGF_H */

