/*

  ssheprintf.c

  Author: Antti Huima <huima@ssh.fi>

  Copyright (c) 1999 SSH Communications Security, Finland
  All rights reserved.

  Created Wed Oct 13 15:16:58 1999.

  */

#define SSH_MAKING_EPRINTF
#undef  NO_SNPRINTF

#include "snprintf.c"

