/*

  sshadt_array.h

  Author: Antti Huima <huima@ssh.fi>

  Copyright (c) 1999 SSH Communications Security, Finland
  All rights reserved.

  Created Tue Oct 19 10:58:11 1999.

  */

#ifndef SSH_ADT_ARRAY_H_INCLUDED
#define SSH_ADT_ARRAY_H_INCLUDED

#include "sshadt.h"

extern const SshADTContainerType ssh_adt_array_type;

#define SSH_ADT_ARRAY (ssh_adt_array_type)

#endif /* SSH_ADT_ARRAY_H_INCLUDED */
